#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "pen.h"
#ifdef HAVE_KQUEUE
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>

static int kq;
static struct kevent *kev;
static int nfds;
static int count;
static int index;

static void kqueue_event_reset(void)
{
	DEBUG(2, "kqueue_event_reset()");
        nfds = 0;
}

static void kqueue_event_arm(int fd, int events)
{
        int kqevents = 0;
	DEBUG(2, "kqueue_event_arm(fd=%d, events=%d)", fd, events);
        if (events & EVENT_READ) kqevents |= EVFILT_READ;
        if (events & EVENT_WRITE) kqevents |= EVFILT_WRITE;
        EV_SET(&kev[nfds], fd, kqevents,
                EV_ADD | EV_ONESHOT, 0, 0, 0);
        nfds++;
}

static void kqueue_event_wait(void)
{
        struct timespec tv;
	DEBUG(2, "kqueue_event_wait()");
        tv.tv_sec = timeout;
        tv.tv_nsec = 0;
	index = -1;
        count = kevent(kq, kev, nfds, kev, nfds, &tv);
	DEBUG(2, "kevent returns %d", count);
        if (count < 0 && errno != EINTR) {
                perror("kevent");
                error("Error on kevent");
        }
}

static int kqueue_event_fd(int *revents)
{
        int events = 0;
	DEBUG(2, "kqueue_event_fd(revents=%p)", revents);
        index++;
        if (index >= count) return -1;
	DEBUG(3, "\tkev[%d] = {filter=%d, ident=%d}", index, kev[index].filter, kev[index].ident);
        if (kev[index].filter & EVFILT_READ) events |= EVENT_READ;
        if (kev[index].filter & EVFILT_WRITE) events |= EVENT_WRITE;
	*revents = events;
        return kev[index].ident;
}

void kqueue_init(void)
{
	kq = kqueue();
	if (kq == -1) {
		perror("kqueue");
		error("Error creating kernel queue");
	}
	kev = pen_malloc((connections_max*2+2)*sizeof *kev);
	event_reset = kqueue_event_reset;
	event_arm = kqueue_event_arm;
	event_wait = kqueue_event_wait;
	event_fd = kqueue_event_fd;
}
#else
void kqueue_init(void)
{
	debug("You don't have kqueue");
	exit(EXIT_FAILURE);
}
#endif
