#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "pen.h"
#ifdef HAVE_POLL
#include <poll.h>

struct pollfd *poll_ufds;
static int poll_nfds, poll_count;
static int index;

static void poll_event_reset(void)
{
	DEBUG(2, "poll_event_reset()");
        poll_nfds = 0;
}

static void poll_event_arm(int fd, int events)
{
        int pollevents = 0;
	DEBUG(2, "poll_event_arm(fd=%d, events=%d)", fd, events);
        if (events & EVENT_READ) pollevents |= POLLIN;
        if (events & EVENT_WRITE) pollevents |= POLLOUT;
        poll_ufds[poll_nfds].fd = fd;
        poll_ufds[poll_nfds].events = pollevents;
        poll_nfds++;
}

static void poll_event_wait(void)
{
	DEBUG(2, "poll_event_wait()");
	index = -1;
        poll_count = poll(poll_ufds, poll_nfds, 1000*timeout);
	DEBUG(2, "poll returns %d", poll_count);
        if (poll_count < 0 && errno != EINTR) {
                perror("poll");
                error("Error on poll");
        }
}

static int poll_event_fd(int *revents)
{
        int events = 0;
	DEBUG(2, "poll_event_fd(revents=%p)", revents);
	for (index++; index < poll_nfds; index++) {
		DEBUG(3, "\tpoll_ufds[%d] = {fd=%d, revents=%d}", index, poll_ufds[index].fd, poll_ufds[index].revents);
        	if (poll_ufds[index].revents & POLLIN) events |= EVENT_READ;
        	if (poll_ufds[index].revents & POLLOUT) events |= EVENT_WRITE;
		if (events) {
			*revents = events;
        		return poll_ufds[index].fd;
		}
	}
	return -1;
}

void poll_init(void)
{
	poll_ufds = pen_malloc((connections_max*2+2)*sizeof *poll_ufds);
	event_reset = poll_event_reset;
	event_arm = poll_event_arm;
	event_wait = poll_event_wait;
	event_fd = poll_event_fd;
}
#else
void poll_init(void)
{
	debug("You don't have poll");
	exit(EXIT_FAILURE);
}
#endif
