#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#if defined HAVE_SYS_SELECT_H && !defined WINDOWS
/* because windows doesn't have it */
#include <sys/select.h>
#endif
#include "pen.h"

static fd_set w_read, w_write, w_error;
static int w_max;

static void select_event_reset(void)
{
	DEBUG(2, "select_event_reset()");
        FD_ZERO(&w_read);
        FD_ZERO(&w_write);
        FD_ZERO(&w_error);
        w_max = 0;
}

static void select_event_arm(int fd, int events)
{
	DEBUG(2, "select_event_arm(fd=%d, events=%d)", fd, events);
        if (events & EVENT_READ) FD_SET(fd, &w_read);
        if (events & EVENT_WRITE) FD_SET(fd, &w_write);
        if (fd >= w_max) w_max = fd+1;
}

static int fd;

static void select_event_wait(void)
{
	int n, err;
        struct timeval tv;
	DEBUG(2, "select_event_wait()");
        tv.tv_sec = timeout;
        tv.tv_usec = 0;
	fd = -1;
        n = select(w_max, &w_read, &w_write, 0, &tv);
	err = socket_errno;
	DEBUG(2, "select returns %d, socket_errno=%d", n, err);
        if (n < 0 && err != EINTR) {
                perror("select");
                error("Error on select");
        }
}

static int select_event_fd(int *revents)
{
        int events = 0;
	DEBUG(2, "select_event_fd(revents=%p)", revents);
	for (fd++; fd < w_max; fd++) {
                if (FD_ISSET(fd, &w_read)) events |= EVENT_READ;
                if (FD_ISSET(fd, &w_write)) events |= EVENT_WRITE;
                if (events) {
			*revents = events;
			return fd;
		}
        }
        return -1;
}

void select_init(void)
{
	if ((connections_max*2+10) > FD_SETSIZE) {
		error("Number of simultaneous connections to large.\n"
			"Maximum is %d, or re-build pen with larger FD_SETSIZE",
			(FD_SETSIZE-10)/2);
	}
	event_reset = select_event_reset;
	event_arm = select_event_arm;
	event_wait = select_event_wait;
	event_fd = select_event_fd;
}

